from requests.exceptions import HTTPError

from solaredge2mqtt.exceptions import ConfigurationException, InvalidDataException
from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import (
    LogicalInfo,
    LogicalInverter,
    LogicalModule,
    LogicalString,
)
from solaredge2mqtt.mqtt import MQTTClient
from solaredge2mqtt.service.http import HTTPClient
from solaredge2mqtt.settings import MonitoringSettings

LOGIN_URL = "https://monitoring.solaredge.com/solaredge-apigw/api/login"
LOGICAL_URL = "https://monitoring.solaredge.com/solaredge-apigw/api/sites/{site_id}/layout/logical"


class MonitoringSite(HTTPClient):
    def __init__(self, settings: MonitoringSettings, mqtt: MQTTClient) -> None:
        super().__init__("Monitoring Site")
        self.settings = settings

        self.mqtt = mqtt

    async def loop(self):
        modules = self.get_module_energies()

        if modules is None:
            logger.warning("Invalid monitoring data, skipping this loop")
            return

        energy_total = 0
        count_modules = 0
        for module in modules:
            if module.energy is not None:
                count_modules += 1
                energy_total += module.energy

        logger.info(
            "Read from monitoring total energy: {energy_total} kWh from {count_modules} modules",
            energy_total=energy_total / 1000,
            count_modules=count_modules,
        )

        await self.mqtt.publish_to("monitoring/pv_energy_today", energy_total)
        for module in modules:
            await self.mqtt.publish_to(
                f"monitoring/module/{module.info.serialnumber}",
                module,
            )

    def login(self) -> None:
        try:
            self._post(
                LOGIN_URL,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data={
                    "j_username": self.settings.username,
                    "j_password": self.settings.password.get_secret_value(),
                },
                timeout=10,
                expect_json=False,
            )

            logger.info("Login to monitoring site successful")
        except HTTPError as error:
            raise ConfigurationException(
                "Monitoring","Unable to login to monitoring account"
            ) from error

    def get_module_energies(self) -> list[LogicalInverter] | None:
        modules = None
        logical = self._get_logical()

        if logical is None:
            raise InvalidDataException(
                "Unable to read logical layout from monitoring site"
            )

        inverters = self._parse_inverters(
            logical["logicalTree"]["children"], logical["reportersData"]
        )

        modules = []

        for inverter in inverters:
            logger.debug(
                "Inverter: {inverter}", inverter=inverter.model_dump_json(indent=4)
            )
            for string in inverter.strings:
                for module in string.modules:
                    modules.append(module)

        return modules

    def _get_logical(self) -> dict | None:
        result = None
        try:
            result = self._get(
                LOGICAL_URL.format(site_id=self.settings.site_id),
                headers={
                    "Content-Type": "application/x-www-form-urlencoded",
                    "X-CSRF-TOKEN": self.session.cookies["CSRF-TOKEN"],
                },
                timeout=10,
            )
        except HTTPError as error:
            raise InvalidDataException("Unable to read logical layout") from error

        return result

    def _parse_inverters(self, inverter_objs, reporters_data) -> list[LogicalInverter]:
        inverters = []

        for inverter_obj in inverter_objs:
            info = LogicalInfo.map(inverter_obj["data"])
            if "INVERTER" in info["type"]:
                inverter = LogicalInverter(
                    info=info,
                    energy=(
                        reporters_data[info["id"]]["unscaledEnergy"]
                        if info["id"] in reporters_data
                        else None
                    ),
                )

                self._parse_strings(inverter, inverter_obj["children"], reporters_data)

                inverters.append(inverter)

            else:
                logger.info("Unknown inverter type: {type}", type=info["type"])

        return inverters

    def _parse_strings(self, inverter, string_objs, reporters_data):
        for string_obj in string_objs:
            info = LogicalInfo.map(string_obj["data"])
            string = LogicalString(
                info=info,
                energy=(
                    reporters_data[info["id"]]["unscaledEnergy"]
                    if info["id"] in reporters_data
                    else None
                ),
            )

            self._parse_panels(string, string_obj["children"], reporters_data)

            inverter.strings.append(string)

    def _parse_panels(self, string, panel_objs, reporters_data):
        for panel_obj in panel_objs:
            info = LogicalInfo.map(panel_obj["data"])
            panel = LogicalModule(
                info=info,
                energy=(
                    reporters_data[info["id"]]["unscaledEnergy"]
                    if info["id"] in reporters_data
                    else None
                ),
            )

            string.modules.append(panel)



from aiomqtt import Client, Will
from pydantic import BaseModel

from solaredge2mqtt.logging import logger
from solaredge2mqtt.settings import MQTTSettings


class MQTTClient(Client):
    def __init__(self, settings: MQTTSettings):
        self.broker = settings.broker
        self.port = settings.port

        self.topic_prefix = settings.topic_prefix

        logger.info(
            "Using MQTT broker: {broker}:{port}",
            broker=settings.broker,
            port=settings.port,
        )

        will = Will(
            topic=f"{self.topic_prefix}/status", payload="offline", qos=1, retain=True
        )

        super().__init__(
            self.broker,
            self.port,
            username=settings.username,
            password=settings.password.get_secret_value(),
            identifier=settings.client_id,
            will=will,
        )

    async def publish_status_online(self) -> None:
        await self.publish_to("status", "online", True)

    async def publish_status_offline(self) -> None:
        await self.publish_to("status", "offline", True)

    async def publish_to(
        self,
        topic: str,
        payload: str | int | float | BaseModel,
        retain: bool = False,
        qos: int = 1,
    ) -> None:
        if self._connected:
            if isinstance(payload, BaseModel):
                payload = payload.model_dump_json()
            await self.publish(
                f"{self.topic_prefix}/{topic}", payload, qos=qos, retain=retain
            )



from datetime import datetime

from requests.exceptions import HTTPError

from solaredge2mqtt.exceptions import InvalidDataException
from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import (
    OpenWeatherMapOneCall,
    OpenWeatherMapOneCallTimemachine,
)
from solaredge2mqtt.mqtt import MQTTClient
from solaredge2mqtt.service.http import HTTPClient
from solaredge2mqtt.settings import ServiceSettings

ONECALL_URL = "https://api.openweathermap.org/data/3.0/onecall"
TIMEMACHINE_URL = "https://api.openweathermap.org/data/3.0/onecall/timemachine"


class WeatherClient(HTTPClient):
    def __init__(self, settings: ServiceSettings, mqtt: MQTTClient) -> None:
        super().__init__("Weather API")

        self.location = settings.location
        self.settings = settings.weather

        self.mqtt = mqtt

    async def loop(self):
        weather = self.get_weather()
        await self.mqtt.publish_to("weather/current", weather.current)

    def get_weather(self) -> OpenWeatherMapOneCall:
        try:
            result = self._get(
                ONECALL_URL,
                params={
                    "lat": self.location.latitude,
                    "lon": self.location.longitude,
                    "exclude": "minutely,daily,alerts",
                    "units": "metric",
                    "lang": self.settings.language,
                    "appid": self.settings.api_key.get_secret_value(),
                },
                timeout=7,
            )

            logger.trace(result)

            if result is None:
                raise InvalidDataException("Unable to read weather data")

            weather = OpenWeatherMapOneCall(**result)

            return weather
        except HTTPError as error:
            raise InvalidDataException("Unable to read weather data") from error

    def get_timemachine(self, dt: datetime) -> OpenWeatherMapOneCall:
        try:
            result = self._get(
                TIMEMACHINE_URL,
                params={
                    "lat": self.location.latitude,
                    "lon": self.location.longitude,
                    "dt": round(dt.timestamp()),
                    "units": "metric",
                    "lang": self.settings.language,
                    "appid": self.settings.api_key.get_secret_value(),
                },
                timeout=10,
            )

            logger.trace(result)

            if result is None:
                raise InvalidDataException("Unable to read timemachine weather data")

            weather = OpenWeatherMapOneCallTimemachine(**result)

            return weather
        except HTTPError as error:
            raise InvalidDataException(
                "Unable to read timemachine weather data"
            ) from error



from datetime import datetime, timedelta, timezone

import pkg_resources
from influxdb_client import BucketRetentionRules, InfluxDBClient, Point
from influxdb_client.client.exceptions import InfluxDBError
from influxdb_client.client.influxdb_client_async import (
    InfluxDBClientAsync,
    QueryApiAsync,
)
from pandas import DataFrame

from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import ForecastPeriod, HistoricPeriod
from solaredge2mqtt.settings import InfluxDBSettings, PriceSettings


class InfluxDB:
    def __init__(self, settings: InfluxDBSettings, prices: PriceSettings) -> None:
        self.settings: InfluxDBSettings = settings
        self.prices: PriceSettings = prices

        self.client: InfluxDBClient = InfluxDBClient(
            url=f"{settings.host}:{settings.port}",
            token=settings.token.get_secret_value(),
            org=settings.org,
        )

        self.write_api = self.client.write_api(
            success_callback=self.write_success_callback,
            error_callback=self.write_error_callback,
            retry_callback=self.write_error_callback,
        )

        self.query_api = self.client.query_api()
        self.delete_api = self.client.delete_api()

        self.client_async: InfluxDBClientAsync | None = None
        self.query_api_async: QueryApiAsync | None = None

        self.flux_cache: dict[str, str] = {}

    def initialize_buckets(self) -> None:
        buckets_api = self.client.buckets_api()
        bucket = buckets_api.find_bucket_by_name(self.bucket_name)
        retention_rules = BucketRetentionRules(
            type="expire", every_seconds=self.settings.retention
        )

        if bucket is None:
            logger.info(f"Creating bucket '{self.bucket_name}'")
            buckets_api.create_bucket(
                bucket_name=self.bucket_name, retention_rules=retention_rules
            )
        else:
            logger.info(f"Bucket '{self.bucket_name}' already exists.")
            if bucket.retention_rules[0].every_seconds != self.settings.retention:
                logger.info(
                    f"Updating retention rules for bucket '{self.bucket_name}' "
                    + f"to {self.settings.retention} seconds."
                )
                bucket.retention_rules[0] = retention_rules
                buckets_api.update_bucket(bucket=bucket)

    async def loop(self) -> None:
        now = datetime.now(tz=timezone.utc).replace(minute=0, second=0, microsecond=0)

        logger.info("Aggregate powerflow and energy raw data")
        aggregate_query = self._get_flux_query(
            "aggregate",
            {"PRICE_IN": self.prices.price_in, "PRICE_OUT": self.prices.price_out},
        )
        self.query_api.query(aggregate_query)

        logger.info("Apply retention on raw data")
        retention_time = now - timedelta(hours=self.settings.retention_raw)
        self.delete_from_measurements(
            datetime(1970, 1, 1, tzinfo=timezone.utc),
            retention_time,
            ["powerflow_raw", "battery_raw"],
        )

    def delete_from_measurements(
        self, start: datetime, stop: datetime, measurements: list[str]
    ) -> None:
        for measurement in measurements:
            self.delete_from_measurement(start, stop, measurement)

    def delete_from_measurement(
        self, start: datetime, stop: datetime, measurement: str
    ) -> None:
        self.delete_api.delete(
            start, stop, f'_measurement="{measurement}"', self.bucket_name
        )

    @property
    def bucket_name(self) -> str:
        return f"{self.settings.bucket}"

    def write_point(self, point: Point) -> None:
        self.write_points([point])

    def write_points(self, points: list[Point]) -> None:
        self.write_api.write(bucket=self.bucket_name, record=points)

    def write_success_callback(self, conf: tuple[str, str, str], data: str) -> None:
        logger.debug(f"InfluxDB batch written: {conf} {data}")

    def write_error_callback(
        self, conf: tuple[str, str, str], data: str, error: InfluxDBError
    ) -> None:
        logger.error(f"InfluxDB error while writting: {conf} {error}")
        logger.debug(data)

    def query_timeunit(
        self, period: HistoricPeriod | ForecastPeriod, measurement: str
    ) -> dict[str, any] | None:
        return self.query_first(
            period.query.query, {"UNIT": period.unit, "MEASUREMENT": measurement}
        )

    def query_first(
        self, query_name: str, additional_replacements: dict[str, any] | None = None
    ) -> dict[str, any] | None:
        result = self.query(query_name, additional_replacements)
        return result[0] if len(result) > 0 else None

    def query(
        self, query_name: str, additional_replacements: dict[str, any] | None = None
    ) -> list[dict[str, any]]:

        tables = self.query_api.query(
            self._get_flux_query(query_name, additional_replacements)
        )
        return [record.values for table in tables for record in table.records]

    async def query_dataframe(
        self, query_name: str, additional_replacements: dict[str, any] | None = None
    ) -> DataFrame:
        await self.init_client_async()
        async with self.client_async:
            return await self.query_api_async.query_data_frame(
                self._get_flux_query(query_name, additional_replacements)
            )

    async def init_client_async(self) -> None:
        self.client_async = InfluxDBClientAsync(
            url=f"{self.settings.host}:{self.settings.port}",
            token=self.settings.token.get_secret_value(),
            org=self.settings.org,
        )
        self.query_api_async = self.client_async.query_api()

    def _get_flux_query(
        self, query_name: str, additional_replacements: dict[str, any] | None = None
    ) -> str:
        if query_name not in self.flux_cache:
            flux = pkg_resources.resource_string(
                __name__, f"../flux/{query_name}.flux"
            ).decode("utf-8")
            flux = (
                flux.replace("{{BUCKET_AGGREGATED}}", self.bucket_name)
                .replace("{{BUCKET_NAME}}", self.bucket_name)
                .replace("{{TIMEZONE}}", self.settings.timezone)
            )
            self.flux_cache[query_name] = flux

        query = self.flux_cache[query_name]

        if additional_replacements is not None:
            for key, value in additional_replacements.items():
                query = query.replace("{{" + key + "}}", str(value))

        logger.trace(query)

        return query
